#ifndef _ASYNC_PROGRESS_H
#define _ASYNC_PROGRESS_H

#include <infiniband/verbs.h>
#include "vbuf.h"

typedef enum {
    ASYNC_QP_DISCONNECTED = 0,
    ASYNC_QP_REQUESTED,
    ASYNC_QP_REPLIED,
    ASYNC_QP_CONNECTED,
    ASYNC_QP_REJECTED
} async_qp_state_t;

typedef struct {
    pthread_spinlock_t             lock;
    vbuf                           *head;
    vbuf                           *tail;
} vbuf_queue_t;

typedef struct _async_prog_conn_t {
    async_qp_state_t          qp_state;
    int                       msgs_sent;
    struct ibv_qp             *qp;
    struct _async_prog_conn_t *next;
    int                       send_wqes_avail;
    struct vbuf *             ext_sendq_head;
    struct vbuf *             ext_sendq_tail;
    int                       ext_sendq_size;
    uint16_t                  next_packet_tosend;
    uint16_t                  next_packet_expected;
} async_prog_conn_t;

typedef struct {
    struct ibv_cq                  *cq;
    struct ibv_comp_channel        *comp_channel;
    struct ibv_srq                 *srq;
    pthread_t                      completion_thread;
    async_prog_conn_t              *head;
    async_prog_conn_t              *tail;
    pthread_spinlock_t             lock;
    vbuf_queue_t                   vbufq;
} async_prog_info_t;

#define ADD_ASYNC_CONN(_i, _c) {                                \
    if((_i).tail) {                                             \
        (_i).tail->next = (_c);                                 \
    }                                                           \
    (_i).tail = (_c);                                           \
    (_c)->next = NULL;                                          \
    if(NULL == (_i).head) {                                     \
        (_i).head = (_c);                                       \
    }                                                           \
}

#define REMOVE_ASYNC_CONN(_i, _ptr_c) {                         \
    if((_i).head) {                                             \
        *(_ptr_c) = (_i).head;                                  \
        (_i).head = (_i).head->next;                            \
    } else {                                                    \
        *(_ptr_c) = NULL;                                       \
    }                                                           \
}

#define VBUF_QUEUE_EMPTY(_q) (NULL == (_q)->head)

#define VBUF_QUEUE_ENQUE(_q, _v) {                              \
    pthread_spin_lock(&((_q)->lock));                           \
    if((_q)->tail) {                                            \
        (_q)->tail->shandle = (_v);                             \
    }                                                           \
    (_q)->tail = (_v);                                          \
    (_v)->shandle = NULL;                                       \
    if(NULL == (_q)->head) {                                    \
        (_q)->head = (_v);                                      \
    }                                                           \
    pthread_spin_unlock(&((_q)->lock));                         \
}

#define VBUF_QUEUE_DEQUEUE(_q, _v_ptr) {                        \
    pthread_spin_lock(&((_q)->lock));                           \
    if((_q)->head) {                                            \
        *(_v_ptr) = (_q)->head;                                 \
        (_q)->head = (_q)->head->shandle;                       \
    } else {                                                    \
        *(_v_ptr) = NULL;                                       \
    }                                                           \
    pthread_spin_unlock(&((_q)->lock));                         \
}

/* Walk through the VBUF queue to see if there are any
 * in order messages that can be processed now.
 * If there are no msgs which are ready to be processed,
 * then return NULL in _v_ptr */
#define INORDER_VBUF_AVAIL(_q, _v_ptr) {                        \
    vbuf* _v;                                                   \
    vbuf* _prev_v;                                              \
    viadev_packet_header *_v_header;                            \
    viadev_connection_t *_conn;                                 \
    packet_sequence_t _seq_id, _seq_id_expected;                \
    *(_v_ptr) = NULL;                                           \
    pthread_spin_lock(&((_q)->lock));                           \
    _v = (_q)->head;                                            \
    _prev_v = (_q)->head;                                       \
    while(_v) {                                                 \
        _v_header = (viadev_packet_header *)                    \
                    VBUF_BUFFER_START((_v));                    \
        _seq_id = (_v_header)->id;                              \
        _conn = &viadev.connections[(_v_header)->src_rank];     \
        _seq_id_expected = (_conn)->next_packet_expected;       \
        if(_seq_id == _seq_id_expected) {                       \
            *(_v_ptr) = _v;                                     \
            /* Dequeue THIS vbuf */                             \
            if( _v == (_q)->head) {                             \
                (_q)->head = (vbuf *) (_v)->shandle;            \
            } else {                                            \
                (_prev_v)->shandle = (_v)->shandle;             \
            }                                                   \
            if(NULL == (_q)->head) {                            \
                (_q)->tail = NULL;                              \
            }                                                   \
            break;                                              \
        }                                                       \
        _prev_v = _v;                                           \
        _v = (vbuf *) ((_v)->shandle);                          \
    }                                                           \
    pthread_spin_unlock(&((_q)->lock));                         \
}

/* Walk through the VBUF queue and find out if
 * a particular connection has an in-order msg
 * or not. Return in pointer _v_ptr */
#define INORDER_VBUF_AVAIL_THIS_CONN(_q, _c, _v_ptr) {          \
    vbuf* _v;                                                   \
    vbuf* _prev_v;                                              \
    viadev_packet_header *_v_header;                            \
    viadev_connection_t *_conn;                                 \
    packet_sequence_t _seq_id, _seq_id_expected;                \
    *(_v_ptr) = NULL;                                           \
    pthread_spin_lock(&((_q)->lock));                           \
    _v = (_q)->head;                                            \
    _prev_v = (_q)->head;                                       \
    while(_v) {                                                 \
        _v_header = (viadev_packet_header *)                    \
                    VBUF_BUFFER_START((_v));                    \
        _seq_id = (_v_header)->id;                              \
        _conn = &viadev.connections[(_v_header)->src_rank];     \
        if(_conn == _c) {                                       \
            _seq_id_expected = (_c)->next_packet_expected;      \
            if(_seq_id == _seq_id_expected) {                   \
                *(_v_ptr) = _v;                                 \
                /* Dequeue THIS vbuf */                         \
                if( _v == (_q)->head) {                         \
                    (_q)->head = (vbuf *) (_v)->shandle;        \
                } else {                                        \
                    (_prev_v)->shandle = (_v)->shandle;         \
                }                                               \
                if(NULL == (_q)->head) {                        \
                    (_q)->tail = NULL;                          \
                }                                               \
                break;                                          \
            }                                                   \
        }                                                       \
        _prev_v = _v;                                           \
        _v = (vbuf *) ((_v)->shandle);                          \
    }                                                           \
    pthread_spin_unlock(&((_q)->lock));                         \
}

#endif /* _ASYNC_PROGRESS_H */
