

#ifndef _MV_PACKET_H
#define _MV_PACKET_H

#include "ibverbs_header.h"
#include "mv_param.h"

/* request ID is generated by sender and sent to receiver. Receiver
 * returns this ID in an ACK in order to allow sender to quickly retrieve
 * the request. Currently, this is just a pointer. In the future, if we
 * make use of immediate data with drma write. This might be to avoid
 * the extra RPUT_FINISH message, or it might be for unreliable
 * via. In the first case, the request type will be 32 bits. In the 
 * second case, it may be even less because we would have to fit
 * a sequence number into immediate data. 
 * The first case will fail on 64-bit machines; the second will
 * fail on all machines -- without a redefinition of request_id_t.
 */

typedef void *request_id_t;
typedef uint16_t packet_sequence_t;

/* sequence number and src is in the immediate field for all non-FP
 * transfers
 */

typedef struct {
    uint16_t has_header;
    packet_sequence_t seqnum;
    uint32_t src_rank;
} mvdev_packet_mheader;

typedef struct {
    uint16_t type;
    uint16_t last_recv; 
} mvdev_packet_header;

typedef struct {
    int context;                /* context id -- identifies communicator */
    int tag;                    /* MPI message tag. */
    int data_length;            /* how many bytes in MPI message (not header).  */
    int src_lrank;              /* to help receiver */
} mvdev_packet_envelope;


typedef enum {
    MVDEV_TRANSPORT_UD = 1
} mvdev_transport;

typedef enum {
    MVDEV_PROTOCOL_EAGER = 1,
    MVDEV_PROTOCOL_R3,

    MVDEV_PROTOCOL_UD_ZCOPY, 

    MVDEV_PROTOCOL_RC_RPUT,
    MVDEV_PROTOCOL_RC_RGET,
    MVDEV_PROTOCOL_RENDEZVOUS_UNSPECIFIED,

    MVDEV_PROTOCOL_SMP_SHORT,
    MVDEV_PROTOCOL_SMP_RNDV,
} mvdev_protocol_t;

typedef struct {
    mvdev_packet_header header;
    mvdev_packet_envelope envelope;
} mvdev_packet_eager_start;

typedef struct {
    mvdev_packet_header header;
} mvdev_packet_eager_next;

/* unfortunately we cannot use the immediate field
 * with RDMA FP without hurting the performance
 */
typedef struct {
    mvdev_packet_header header;
    mvdev_packet_envelope envelope;
    packet_sequence_t seqnum;
} mvdev_packet_fp_eager_start;

typedef struct {
    mvdev_packet_header header; 
    mvdev_packet_envelope envelope;
    request_id_t sreq;
    int protocol;
} mvdev_packet_rendezvous_start;

typedef struct {
    mvdev_packet_header header; 
    request_id_t rreq;          /* identifies receive handle */
    request_id_t sreq;          /* identifies send handle (echoed) */
    int protocol;               /* explicitly specify protocol */

    uint32_t rndv_qpn;
    uint32_t start_seq;
} mvdev_packet_rendezvous_reply;

typedef struct {
    mvdev_packet_header header;
    request_id_t rreq;
} mvdev_packet_ud_zcopy_finish;

typedef struct {
    mvdev_packet_header header;
    request_id_t sreq;
} mvdev_packet_ud_zcopy_ack;

typedef struct {
    mvdev_packet_header header;
} mvdev_packet_basic;

#define mvdev_packet_ack        mvdev_packet_basic


typedef struct {
    mvdev_packet_header header;
    request_id_t rreq;
} mvdev_packet_r3_data;

typedef enum {
    MVDEV_PACKET_EAGER_START = 0,
    MVDEV_PACKET_EAGER_NEXT,
    MVDEV_PACKET_FP_EAGER_START,

    MVDEV_PACKET_RENDEZVOUS_START,
    MVDEV_PACKET_RENDEZVOUS_REPLY,

    MVDEV_PACKET_R3_DATA,

    MVDEV_PACKET_UD_ZCOPY_FINISH,
    MVDEV_PACKET_UD_ZCOPY_ACK,
    MVDEV_PACKET_UD_ZCOPY_NAK,

    MVDEV_PACKET_ACK,
} mvdev_packet_type;



#define PACKET_SET_HEADER(p, c, t) {                               \
        p->header.type = t;                                        \
        p->header.last_recv = c->seqnum_next_toack;                \
        c->ud_remote_credit = 0; \
}

#define PACKET_SET_ENVELOPE(p, ctx, tag, len, src) {                \
        p->envelope.context     = ctx;                                  \
        p->envelope.tag         = tag;                                  \
        p->envelope.data_length = len;                                  \
        p->envelope.src_lrank   = src;                                  \
}

#define PACKET_SET_CREDITS(p, c) {                                  \
        p->vbuf_credit = c->local_credit;                               \
        c->local_credit = 0;                                            \
        p->remote_credit = c->remote_credit;                            \
}


#endif
