/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */

/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include "mpid.h"
#include "ibverbs_header.h"
#include "mv_dev.h"
#include "mv_priv.h"
#include "mv_inline.h"
#include "mv_buf.h"
#include "mv_packet.h"

mvdev_info_t mvdev;
mv_func mvdev_func;
char nullsbuffer, nullrbuffer;

#define SEND_WIN_CHECK(_c, _v, _len) {                                                      \
    if(MVDEV_UNLIKELY((c->send_window_segments > mvparams.send_window_segments_max ) ||     \
                (c->ext_window_head != NULL && c->ext_window_head != v))) {                 \
        mvdev_windowq_queue(c, v, _len);                                                    \
        return;                                                                             \
    }                                                                                       \
}                                                                                           

#define GET_UD_QP(_toset, _passed) {                                                        \
    if(MVDEV_LIKELY(NULL == _passed)) {                                                     \
        mvdev.last_qp = (mvdev.last_qp + 1) % mvdev.num_ud_qps;                             \
        _toset = &(mvdev.ud_qp[mvdev.last_qp]);                                             \
    }                                                                                       \
}

#define SEND_UD_SR(_qp, _v, _segment) {                                                     \
    if(MVDEV_UNLIKELY(_qp->send_wqes_avail <= 0 || (NULL != _qp->ext_sendq_head))) {        \
        mvdev_ext_sendq_queue(_qp, &(_v->desc[_segment]));                                  \
    } else {                                                                                \
        struct ibv_send_wr *bad_wr;                                                         \
        _qp->send_wqes_avail--;                                                             \
        if(MVDEV_UNLIKELY(ibv_post_send(_qp->qp, &(_v->desc[_segment].sr), &bad_wr))) {     \
            error_abort_all(IBV_RETURN_ERR,"Error posting to UD QP %d\n",                   \
                    _qp->send_wqes_avail);                                                  \
        }                                                                                   \
    }                                                                                       \
}


void mvdev_post_channel_send_qp_aux_headers(
        mvdev_connection_t * c, mv_sbuf * v, int total_len, 
        mv_qp *send_qp, int signalled, int lid) {

    mvdev_packet_mheader * mp = (mvdev_packet_mheader *) v->base_ptr;
    mvdev_packet_header * p = (mvdev_packet_header *) v->header_ptr;
    mv_qp *qp = send_qp;

    assert(total_len <= mvparams.mtu - 40);

    SEND_WIN_CHECK(c, v, total_len);

    total_len += MV_HEADER_OFFSET; 
    v->seqnum = v->seqnum_last = c->seqnum_next_tosend;
    v->rank = c->global_rank;
    v->segments = v->left_to_send = 1;

    INCREMENT_SEQ_NUM(&(c->seqnum_next_tosend));
    SET_MHEADER(mp, v->seqnum, 1);

    mvdev.connection_ack_required[c->global_rank] = 0;

    GET_UD_QP(qp, send_qp);
    prepare_ud_descriptor_noheaders( &(v->desc[0]), v->base_ptr, 
            v, total_len, c->global_rank, v->seqnum, 
            1, qp, signalled, lid);

    SEND_UD_SR(qp, v, 0);
    mvdev_track_send(c, v);

    if(MVDEV_UNLIKELY(MVDEV_PACKET_R3_DATA == p->type)) {
        c->r3_segments++;
    }
}

void mvdev_post_channel_send_qp_aux(mvdev_connection_t * c, mv_sbuf * v, 
        int total_len, mv_qp *send_qp, int signaled, int lid) {

    int len = 0, seqnum, bytes_sent = 0;
    unsigned char * current_pos = v->base_ptr;
    mvdev_packet_header * p = (mvdev_packet_header *) v->header_ptr;;
    mv_qp *qp = send_qp;

    D_PRINT("Entering send_qp. destrank: %d, total_len: %d\n",
            c->global_rank, total_len);

    SEND_WIN_CHECK(c, v, total_len);

    v->seqnum = c->seqnum_next_tosend;
    v->rank = c->global_rank;
    v->segments = 0;
    v->left_to_send = 0;

    mvdev.connection_ack_required[c->global_rank] = 0;

    do {
        seqnum = v->seqnum_last = c->seqnum_next_tosend;
        INCREMENT_SEQ_NUM(&(c->seqnum_next_tosend));

        D_PRINT("seqnum to peer %d is %d, segment: %d\n", 
                c->global_rank, seqnum, v->segments);

        len = MIN(mvparams.mtu - 40, total_len - bytes_sent);

        GET_UD_QP(qp, send_qp);
        prepare_ud_descriptor( &(v->desc[v->segments]), 
                current_pos, v, len, c->global_rank, seqnum, 
                (v->segments == 0), qp, signaled, lid);

        SEND_UD_SR(qp, v, v->segments);
        mvdev_track_send(c, v);

        v->segments++;
        v->left_to_send++;

        current_pos += len;
        bytes_sent += len;
    } while(bytes_sent < total_len);

    if(MVDEV_UNLIKELY(MVDEV_PACKET_R3_DATA == p->type)) {
        c->r3_segments += v->segments;
    }
}


struct ibv_mr* register_memory(int hca_num, void* buf, int len)
{
    return (ibv_reg_mr(mvdev.hca[hca_num].pd, buf, len,
                IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE));
}

int deregister_memory(struct ibv_mr* mr)
{
    return (ibv_dereg_mr(mr));
}

void flush_all_messages() {
    /* need to put the necessary code here */
}

static inline void prepare_ud_recv(mv_rbuf *v) {

    v->desc.rr.next = NULL;
    v->desc.rr.wr_id = (aint_t) &(v->desc);
    v->desc.rr.num_sge = 1;
    v->desc.rr.sg_list = &(v->desc.sg_entry);

    v->desc.sg_entry.addr = (uintptr_t) (v->base_ptr);
    v->desc.sg_entry.length = v->region->size_ptr->alloc_size;
    v->desc.sg_entry.lkey = v->region->mem_handle[0]->lkey;

    v->transport = MVDEV_TRANSPORT_UD;
}

int mvdev_post_srq_buffers(int num_bufs)
{
    int i = 0, total = 1;
    mv_rbuf *v, *first_v, *last_v;
    struct ibv_recv_wr *bad_wr;

    D_PRINT("Attempting to post %d recv bufs\n", num_bufs);

    if(num_bufs > mvparams.recvq_size) {
        num_bufs = mvparams.recvq_size;
    }

    first_v = last_v = get_mv_rbuf(mvparams.mtu);
    prepare_ud_recv(first_v);

    for(i = 1; i < num_bufs; i++) {
        total++;
        v = get_mv_rbuf(mvparams.mtu);
        prepare_ud_recv(v);

        last_v->desc.rr.next = &v->desc.rr;
        last_v = v;
    }

    if(ibv_post_recv(mvdev.ud_qp[0].qp,
                &first_v->desc.rr, &bad_wr)) {
        return 0;
    }       

    D_PRINT("Posted %d recvs\n", i);

    return total;
}

