/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */


#include "mpid_bind.h"
#include "mv_dev.h"
#include "mv_priv.h"
#include "queue.h"
#include "mv_buf.h"
#include "mv_param.h"
#include "mv_inline.h"
#include "mpid_smpi.h"
#include "dreg.h"
#include <math.h>


static inline void mvdev_incoming_eager_start(mv_rbuf * v, 
        mvdev_connection_t * c, mvdev_packet_eager_start * header) 
{
    MPIR_RHANDLE *rhandle;
    int found;

    MPID_Msg_arrived(header->envelope.src_lrank,
                     header->envelope.tag,
                     header->envelope.context, &rhandle, &found);

    /* This information needs to be filled in whether a preposted 
     * receive was found or not. 
     */
    rhandle->connection = c;
    rhandle->s.count = header->envelope.data_length;
    rhandle->protocol = MVDEV_PROTOCOL_EAGER;
    rhandle->vbufs_received = 1;


    /* recv is already posted */
    if (found) {
        int truncated = 0; 
        int copy_bytes;

        /* matching packet is here, can no longer cancel recv */
        rhandle->can_cancel = 0;

        /* what if we get a message longer than what was posted.. */
        if(header->envelope.data_length > rhandle->len) {
            truncated = 1;
        } else {
            rhandle->len = header->envelope.data_length;
        }

        if(0 < (copy_bytes = MIN(rhandle->len, v->byte_len_data))) {
            memcpy((char *) (rhandle->buf), v->data_ptr, copy_bytes);
        }

        rhandle->bytes_copied_to_user = copy_bytes;

        if(rhandle->len == rhandle->bytes_copied_to_user) {
            RECV_COMPLETE(rhandle);
            rhandle->s.MPI_ERROR = truncated ?  MPI_ERR_TRUNCATE : MPI_SUCCESS; 
        } else {
            c->rhandle = rhandle;
            rhandle->s.MPI_ERROR = truncated ?  MPI_ERR_TRUNCATE : MPI_SUCCESS;
        }

        release_mv_rbuf(v);
    } 

    /* Recv not yet posted, set size of incoming data in
     * unexpected queue rhandle
     */
    else {
        rhandle->len = header->envelope.data_length;
        rhandle->vbuf_head = rhandle->vbuf_tail = v;
        rhandle->bytes_copied_to_user = 0;
        v->eager_ptr.next = NULL;
        c->rhandle = rhandle; 

        D_PRINT("unexpected, srcrank: %d, tag: %d, context: %d \n",
                header->envelope.src_lrank, header->envelope.tag, 
                header->envelope.context);
    }

}

static inline void mvdev_incoming_eager_next(mv_rbuf * v, mvdev_connection_t * c)
{
    MPIR_RHANDLE *rhandle = c->rhandle;
    int copy_bytes;

    if(rhandle->bytes_copied_to_user > 0) {
        copy_bytes = MIN(v->byte_len_data,
                (rhandle->len - rhandle->bytes_copied_to_user));
          
        if(copy_bytes) {
            memcpy((char *) (rhandle->buf) +
                    rhandle->bytes_copied_to_user, v->data_ptr, copy_bytes);
            rhandle->bytes_copied_to_user += copy_bytes;
        }

        if(rhandle->s.count == rhandle->bytes_copied_to_user) {
            int err = rhandle->s.MPI_ERROR;
            RECV_COMPLETE(rhandle);
            rhandle->s.MPI_ERROR = err;
        }

        release_mv_rbuf(v);
    } else {
        ((mv_rbuf *) rhandle->vbuf_tail)->eager_ptr.next = v;
        v->eager_ptr.next = NULL;
        rhandle->vbuf_tail = v;
    }
}

static inline void mvdev_incoming_r3_data(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_r3_data * h)
{
    MPIR_RHANDLE *rhandle = (MPIR_RHANDLE *) ID_TO_REQ(h->rreq);

    memcpy(((char *) rhandle->buf) +
           rhandle->bytes_copied_to_user, v->data_ptr, v->byte_len_data);
    rhandle->bytes_copied_to_user += v->byte_len_data;

    assert(rhandle->bytes_copied_to_user <= rhandle->len);

    if(rhandle->bytes_copied_to_user == rhandle->len) {
        RECV_COMPLETE(rhandle);
    } else {
        c->rhandle = rhandle;
    }
}

static inline void mvdev_incoming_r3_data_next(mv_rbuf * v, 
        mvdev_connection_t * c)
{
    MPIR_RHANDLE *rhandle = (MPIR_RHANDLE *) c->rhandle;;

    memcpy(((char *) rhandle->buf) + rhandle->bytes_copied_to_user, 
            v->data_ptr, v->byte_len_data);

    rhandle->bytes_copied_to_user += v->byte_len_data;
    assert(rhandle->bytes_copied_to_user <= rhandle->len);
    if(rhandle->bytes_copied_to_user == rhandle->len) {
        RECV_COMPLETE(rhandle);
        D_PRINT("R3 recv complete from rank %d total %d",
                c->global_rank, rhandle->len);
    }
}



/* By the time we get here we know the message is in order 
 */

void mvdev_process_recv_inorder(mvdev_connection_t * c, mv_rbuf * v) {

    if(v->has_header) {
        mvdev_packet_header *p = (mvdev_packet_header *) v->header_ptr;

        switch(p->type) {
            case MVDEV_PACKET_EAGER_START:
                mvdev_incoming_eager_start(v, c, 
                        (mvdev_packet_eager_start *) p);
                /* don't release since the message may be unexp */
                break;  
            case MVDEV_PACKET_EAGER_NEXT:
                mvdev_incoming_eager_next(v, c);
                /* don't release since the message may be unexp */
                break;  
            case MVDEV_PACKET_RENDEZVOUS_START:
                mvdev_incoming_rendezvous_start(v, c,
                        (mvdev_packet_rendezvous_start *) p);
                release_mv_rbuf(v);
                break;  
            case MVDEV_PACKET_RENDEZVOUS_REPLY:
                mvdev_incoming_rendezvous_reply(v, c,
                        (mvdev_packet_rendezvous_reply *) p);
                release_mv_rbuf(v);
                break;  
            case MVDEV_PACKET_R3_DATA:
                mvdev_incoming_r3_data(v, c, 
                        (mvdev_packet_r3_data *) p);
                release_mv_rbuf(v);
                break;  
            case MVDEV_PACKET_UD_ZCOPY_FINISH:
                mvdev_incoming_ud_zcopy_finish(v, c, 
                        (mvdev_packet_ud_zcopy_finish *) p);
                release_mv_rbuf(v);
                break;
            case MVDEV_PACKET_UD_ZCOPY_ACK:
                mvdev_incoming_ud_zcopy_ack(v, c,
                        (mvdev_packet_ud_zcopy_ack *) p);
                release_mv_rbuf(v);
                break;
            default:
                error_abort_all(IBV_STATUS_ERR,
                        "Unknown packet type %d in mvdev_process_recv_inorder",
                        p->type);
        }

    } else {
        /* There is no header on this packet. We do know that this is 
         * in order though, so it must go with the last message that came
         * in
         */
        mvdev_protocol_t proto = ((MPIR_RHANDLE *) c->rhandle)->protocol;
        switch(proto) {
            case MVDEV_PROTOCOL_EAGER:
                mvdev_incoming_eager_next(v, c);
                break;
            case MVDEV_PROTOCOL_R3:
                mvdev_incoming_r3_data_next(v, c);
                release_mv_rbuf(v);
                break;
            default:
                error_abort_all(IBV_STATUS_ERR,
                        "Unknown protocol type %d in mvdev_process_recv_inorder",
                        proto);
        }
    }
}



void mvdev_incoming_rendezvous_start(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_rendezvous_start * header) 
{
    MPIR_RHANDLE *rhandle;
    int found;
    MPID_Msg_arrived(header->envelope.src_lrank,
            header->envelope.tag,
            header->envelope.context, &rhandle, &found);

    /* This information needs to be filled in whether a preposted 
     * receive was found or not.  */

    rhandle->dreg_entry = NULL;
    rhandle->connection = c;
    rhandle->send_id = header->sreq;
    rhandle->s.count = header->envelope.data_length;
    rhandle->vbufs_received = 0;
    rhandle->vbuf_head = NULL;
    rhandle->vbuf_tail = NULL;
    rhandle->replied = 0;
    rhandle->protocol = header->protocol;
    rhandle->bytes_copied_to_user = 0;

    if (!found) {
        /* recv not yet posted, set size of incoming data in
         * unexpected queue rhandle */
        rhandle->len = header->envelope.data_length;
    } else {
        /* recv posted already, check if recv buffer is large enough */
        if (header->envelope.data_length > rhandle->len) {
            error_abort_all(IBV_STATUS_ERR,
                    "message truncated. ask %d got %d",
                    rhandle->len, header->envelope.data_length);
        }
        rhandle->len = header->envelope.data_length;

        /* packet now arrived, can no-longer cancel recv */
        rhandle->can_cancel = 0;

        /* can we match sender's protocol? */
        switch (header->protocol) {
            case MVDEV_PROTOCOL_R3:
                mvdev_recv_r3(rhandle);
                break;
            case MVDEV_PROTOCOL_UD_ZCOPY:
                mvdev_recv_ud_zcopy(rhandle);
                break;
            default:
                error_abort_all(IBV_STATUS_ERR, 
                        "Invalid protocol %d\n", rhandle->protocol);
        }
    }
}



void mvdev_incoming_rendezvous_reply(mv_rbuf * v, mvdev_connection_t * c,
        mvdev_packet_rendezvous_reply * header)
{
    MPIR_SHANDLE *s = (MPIR_SHANDLE *) (ID_TO_REQ(header->sreq));

    if(s == NULL) {
        error_abort_all(GEN_ASSERT_ERR, 
                "s == NULL, send handler at mvdev_incoming_rendezvous_reply");
    }

    s->bytes_sent = 0;
    s->receive_id = header->rreq;
    s->nearly_complete = 0;

    switch (s->protocol) {
        case MVDEV_PROTOCOL_R3:
            break;
        case MVDEV_PROTOCOL_UD_ZCOPY:
            if(header->protocol == MVDEV_PROTOCOL_R3) {
                if(s->dreg_entry != NULL) {
                    dreg_unregister(s->dreg_entry);
                   s->dreg_entry = NULL;
                }
                s->protocol = MVDEV_PROTOCOL_R3;
            } else {
                assert(header->protocol == MVDEV_PROTOCOL_UD_ZCOPY);
                s->remote_qpn = header->rndv_qpn;
                s->seqnum = header->start_seq;
                D_PRINT("got reply: remote_qpn: %u, seq: %u\n",
                        s->remote_qpn, s->seqnum);
                /* TODO: fix this */
                s->hca_index = 0;
            }
            break;
        default:
            error_abort_all(IBV_STATUS_ERR,
                    "Invalid protocol %d\n", s->protocol);
    }

    RENDEZVOUS_IN_PROGRESS(c, s);
    PUSH_FLOWLIST(c);

    D_PRINT("push onto flow list\n");

}

inline static void mvdev_rendezvous_push_r3(MPIR_SHANDLE * s, mvdev_connection_t *c) {
    mv_sbuf *v;
    mvdev_packet_r3_data *h;
    int datamax;
    int bytes_in_this_packet;

    /* if we get here, then protocol is R3 */
    while (s->nearly_complete == 0 && 
            c->send_window_segments <= mvparams.send_window_segments_max) {

        v = get_sbuf(c, s->bytes_total - s->bytes_sent + 
                sizeof(mvdev_packet_r3_data));
        v->rank = c->global_rank;

        h = (mvdev_packet_r3_data *) v->header_ptr;
        PACKET_SET_HEADER(h, c, MVDEV_PACKET_R3_DATA);
        h->rreq = s->receive_id;

        /* figure out how much data to send in this packet */
        datamax = SBUF_MAX_DATA_SZ_HDR(v, mvdev_packet_r3_data);

        bytes_in_this_packet = s->bytes_total - s->bytes_sent;
        if (datamax < bytes_in_this_packet) {
            bytes_in_this_packet = datamax;
        }

        /* copy next chunk of data in to vbuf */
        memcpy(((char *) h) + sizeof(mvdev_packet_r3_data), 
                ((char *) s->local_address) +
                s->bytes_sent, bytes_in_this_packet);

        mvdev_post_channel_send(c, v, 
                bytes_in_this_packet + sizeof(mvdev_packet_r3_data));

        s->bytes_sent += bytes_in_this_packet;
        if (s->bytes_sent == s->bytes_total) {
            v->shandle = s;
            s->nearly_complete = 1;
        }
    }

    if (s->nearly_complete) {
        if(s->blocked) {
            c->blocked_count--;
            s->blocked = 0;
        }
    } else {
        if(!s->blocked) {
            c->blocked_count++;
        } else {
            s->blocked = 1;
        }
    }
}

inline static void mvdev_rendezvous_push(MPIR_SHANDLE * s)
{
    mvdev_connection_t *c = (mvdev_connection_t *) s->connection;

    switch(s->protocol) {
        case MVDEV_PROTOCOL_UD_ZCOPY:
            mvdev_rendezvous_push_zcopy(s, c);
            break;
        case MVDEV_PROTOCOL_R3:
            mvdev_rendezvous_push_r3(s, c);
            break;
        default:
            error_abort_all(IBV_STATUS_ERR,
                    "Invalid protocol %d\n", s->protocol);

    }
}


void process_flowlist(void)
{
    MPIR_SHANDLE *s;

    while (flowlist) {
        s = flowlist->shandle_head;
        while (s != NULL) {
            mvdev_rendezvous_push(s);
            if (!s->nearly_complete) {
                break;
            }

            RENDEZVOUS_DONE(flowlist);

            if (s != NULL && s->protocol == MVDEV_PROTOCOL_R3) { 
                SEND_COMPLETE(s);
            } 

            s = flowlist->shandle_head;
        }
        /* now move on to the next connection */
        POP_FLOWLIST();
    }

}

